/*
 * alternative Loesung zu Blatt 11 ...
 *
 * ... unter Verwendung von
 * getopt(), was aber ausdruecklich nicht erlaubt war ;-);
 * vielleicht mal ganz nett, um getopt() kennen zu lernen
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>

char *opt_i = NULL;	/* Option -i gesetzt? => Argument */
int inode;
char *opt_n = NULL;	/* Option -n gesetzt? => Argument */
char *opt_a = NULL;	/* Option -a gesetzt? => Argument */
int atime;
int opt_r = 0;		/* Option -r gesetzt? */
int opt_w = 0;		/* Option -w gesetzt? */
int opt_x = 0;		/* Option -x gesetzt? */
int opt_l = 0;		/* Option -l gesetzt? */

/*
 * Wandelt String s in eine Integer um,
 * die in *val geschrieben wird.
 * Liefert im Erfolgsfall true (!= 0) und
 * sonst false (0).
 */
int to_int(char *s, int *val) {
    char *p;
    errno = 0;	/* errno zuerst auf 0 setzen */
    /*
     * liefert die Zahl, die der String s enthaelt;
     * p zeigt danach auf die Stelle im String nach
     * dieser Zahl
     */
    *val = strtol(s, &p, 10);
    /*
     * Fehler, wenn errno != 0 ist oder aber
     * p noch nicht auf das Stringende (also Null-Byte) zeigt.
     */
    if (errno != 0 || *p != '\0')
	return 0;	/* FEHLER */
    return 1;		/* alles ok */
}

/*
 * In allen Unterverzeichnissen von path rekursiv suchen.
 */
void recurse_subdirs(char *path) {
    DIR *dir;
    struct dirent *entry;

    /* Verbindung zum Verzeichnis oeffnen */
    if (!(dir = opendir(path))) {
	char err[1024];
	sprintf(err, "opendir: %s", path);
	perror(err);
	return;
    }

    /* alle Verzeichniseintraege durchlaufen ... */
    while ((entry = readdir(dir))) {
	char buf[1024];
	struct stat statbuf;

	/* bei . und .. machen wir nichts und gehen gleich zum
	 * naechsten Verzeichniseintrag */
	if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, ".."))
	    continue;

	/* Pfadnamen zusammenbauen aus Pfad + Dateiname */
	sprintf(buf, "%s/%s", path, entry->d_name);

	/* je nachdem, ob -l gesetzt bzw. nicht gesetzt,
	 * verwenden wir stat() bzw. lstat() */
	if (opt_l) {
	    if (stat(buf, &statbuf) < 0) {
		char err[1024];
		sprintf(err, "stat: %s", buf);
		perror(err);
		continue;
	    }
	}
	else {
	    if (lstat(buf, &statbuf) < 0) {
		char err[1024];
		sprintf(err, "lstat: %s", buf);
		perror(err);
		continue;
	    }
	}

	/* fuer die Ausgabe einer Datei muss folgendes fuer jede
	 * Option gelten: Entweder die Option ist nicht angegeben ;-),
	 * oder die Bedingung, die zu dieser Option gehoert, ist
	 * fuer diese Datei erfuellt ... und dies verknuepfen wird
	 * mit UND fuer alle Optionen */
	if ((!opt_i || statbuf.st_ino == inode)			/* Inode-Nummer = inode */
	    && (!opt_n || !strcmp(entry->d_name, opt_n))	/* Dateiname = opt_n (Optionsarg.) */
	    && (!opt_a || atime < statbuf.st_atime)		/* Zugriffszeit neuer als atime */
	    && (!opt_r || (statbuf.st_mode & 0400) != 0)	/* Besitzer der Datei kann lesen */
	    && (!opt_w || (statbuf.st_mode & 0200) != 0)	/* Besitzer der Datei kann schreiben */
	    && (!opt_x || (statbuf.st_mode & 0100) != 0))	/* Besitzer der Datei kann ausfuehren */
	    puts(buf);

	/* ... wenn's sich um ein Verzeichnis handelt
	 * (aber ja nicht mehr . und ..), dann steigen
	 * wir rekursiv hinunter ... */
	if (S_ISDIR(statbuf.st_mode))
	    recurse_subdirs(buf);
    }

    /* ... und die Verbindung zum Verzeichnis sauber beenden */
    closedir(dir);
}

int main(int argc, char *argv[]) {
    int errflag = 0;		/* Fehler bei den Optionen? */
    char c;
    int args;

    /*
     * verarbeite naechste Option;
     * beginne bei Index optind in argv;
     * Optionen sind x, n und a (dieses mit Parameter => ":" nach "a");
     * liefert Optionszeichen bzw. ? bei Fehler und EOF, beim Ende der
     * Optionsverarbeitung
     * (durch Setzen von opterr = 0 koennen die Fehlermeldungen von getopt()
     *  ausgeschaltet werden; im Fehlerfall - also "?" - enthaelt optopt das
     *  Zeichen, das den Fehler ausloeste)
     */
    while ((c = getopt(argc, argv, "i:n:a:rwxl")) != EOF)
	switch (c) {	/* c enthaelt die Option bzw. ? im Fehlerfall */
	    case 'i' : opt_i = optarg; break;	/* optarg ist char * auf den Optionsparameter */
	    case 'n' : opt_n = optarg; break;	/* optarg ist char * auf den Optionsparameter */
	    case 'a' : opt_a = optarg; break;	/* optarg ist char * auf den Optionsparameter */
	    case 'r' : opt_r = 1; break;
	    case 'w' : opt_w = 1; break;
	    case 'x' : opt_x = 1; break;
	    case 'l' : opt_l = 1; break;
	    case '?' : errflag++; break;
	}

    args = argc - optind;	/* Anzahl verbleibender Argumente */

    /* Fehler bei zuvielen Argumenten ausgeben */
    if (args > 1)
	fprintf(stderr, "Too many arguments\n");

    /* Usage-Meldung, falls zuviele Argumente, oder Fehler bei den Optionen */
    if (args > 1 || errflag) {
	fprintf(stderr, "Usage: %s [-i <inode>] [-n <file>] [-a <file>] [-r|-w|-x|-l] [<directory>]\n", argv[0]);
	return 1;
    }

    /* falls Option -i, dann noch den String opt_i parsen => muss Integer >= 0 sein! */
    if (opt_i && (!to_int(opt_i, &inode) || inode < 0)) {
	fprintf(stderr, "Argument of -i is not a non-negative integer: %s\n", opt_i);
	return 2;
    }

    /* Option -a => Zugriffszeit der Datei ermitteln */
    if (opt_a) {
	struct stat statbuf;
	/* meine Wahl: lstat(); stat() waere auch okay */
	if (lstat(opt_a, &statbuf) < 0) {
	    char err[1024];
	    sprintf(err, "lstat: %s", opt_a);
	    perror(err);
	    return 3;
	}
	/* Zugriffszeit der Datei mit dem Namen, der bei -a als Argument angegeben wurde */
	atime = (int) statbuf.st_atime;
    }

    /* jetzt suchen wir rekrusiv in den Unterverzeichnissen
     * von ., falls kein Argument, und sonst im angegebenen Verzeichnis */
    recurse_subdirs(args > 0 ? argv[optind] : ".");

    return 0;
}
